/****************************************************************************
 *  
 * Project: Olimex AVR ISP with MKII protocol support.
 *
 * Copyright by Olimex Ltd. All rights reserved.
 * 
 * File: massbulk.c
 * Description: USB Mass Storage Bulk-Only Module
 * Developer: Dimitar Dimitrov ( dinuxbg __on__  gmail.com )
 *
 * Last change: $Date: 2007-12-12 17:31:49 +0200 (сряда, 12 Декември 2007) $
 * Revision: $Revision: 89 $
 * Id: $Id: massbulk.c 89 2007-12-12 15:31:49Z dimitar $
 *
 ****************************************************************************/

#include "usb.h"
#include "massbulk.h"
#include "usbscsi.h"

#if USB_EP_MS_OUT_MAX_PACKET_SIZE < 32
  #error The MassStorage OUT max packet size must be at least 32!
#endif
#if USB_EP_MS_IN_MAX_PACKET_SIZE < 32
  #error The MassStorage IN max packet size must be at least 32!
#endif


static void ms_stallclr_data_eps(void);

/**
 *  Handle Mass-Storage specific control requests.
 *
 *  @param stppkt USB SETUP packet with class-specific request.
 *
 *  @return Zero value if request was recognized and successfully executed,
 *  non-zero value otherwise.
 */
int8_t ms_usbcb_setupreq(const struct usb_setup_packet *stppkt)
{
	/* sanity check */
	if((stppkt->request_type.type != USB_REQUEST_TYPE_CLASS)
		|| (stppkt->request_type.recipient != USB_RECIPIENT_INTERFACE)
		|| (stppkt->index != MS_INTRFN)) 
	{
		return 1;
	}

	switch(stppkt->request) {
		case MS_CLSREQ_MASS_STORAGE_RESET:
			if((stppkt->value==0) && (stppkt->length==0))
			{
				ms_stallclr_data_eps();
				usb_ep_select(USB_EPN_CONTROL);
				usb_setupreq_sendack();
				usb_ctrlin_send();
				return 0;
			} else {
				return -1;
			}
			break;
		case MS_CLSREQ_GET_MAX_LUN:
			if((stppkt->value==0) && (stppkt->length==1)
					&& stppkt->request_type.dir_d2h)
			{
				usb_setupreq_sendack();
				usb_epfifo_push(MS_MAX_LUN);
				usb_ctrlin_send();
				return 0;
			} else {
				return -1;
			}
			break;
		default:
			return -1;
	}
}

/** Configure data endpoints used by the mass storage driver. */
void ms_usbcb_set_configuration(void)
{
	usb_ep_select(USB_EPN_MS_IN);
	usb_ep_reset(USB_EPN_MS_IN);
	usb_ep_reset_datatoggle();
	if(usb_ep_activate(
			USB_EPN_MS_IN,
			USB_EPTYPE_BULK,
			USB_EPDIR_IN,
			USB_EP_MS_IN_MAX_PACKET_SIZE,
			0	/* no doublebank */
		))
	{
		while(1) ;
	}

	usb_ep_select(USB_EPN_MS_OUT);
	usb_ep_reset(USB_EPN_MS_OUT);
	usb_ep_reset_datatoggle();
	if(usb_ep_activate(
			USB_EPN_MS_OUT,
			USB_EPTYPE_BULK,
			USB_EPDIR_OUT,
			USB_EP_MS_OUT_MAX_PACKET_SIZE,
			0	/* no doublebank */
		))
	{
		while(1) ;
	}
}

/** Stall Mass Storage data endpoints */
static void ms_stall_data_eps(void)
{
	usb_ep_select(USB_EPN_MS_IN);
	usb_ep_stall_enable();
	usb_ep_select(USB_EPN_MS_OUT);
	usb_ep_stall_enable();
}

/** Clear Stall for  Mass Storage data endpoints */
static void ms_stallclr_data_eps(void)
{
	usb_ep_select(USB_EPN_MS_IN);
	usb_ep_stall_disable();
	usb_in_nakclr();
	usb_ep_select(USB_EPN_MS_OUT);
	usb_ep_stall_disable();
}

/** 
 * Fetch a Mass Storage Control Block Wrapper and check it.
 *
 * @param w Where to store the fetched CBW.
 * @return Zero value on success, non-zero value on incorrect CBW length,
 * parameter or signature.
 */
static int8_t ms_recv_cbw(struct ms_cbw *w)
{
	uint8_t *p = (uint8_t*)w;
	uint8_t i = sizeof(*w);

	if(usb_epfifo_bytecount() != sizeof(*w)) {
		return 1;
	}
	while(i) {
		--i;
		*p++ = usb_epfifo_pop();
	}
	usb_out_sendack();
	usb_out_clrbank();

	if(w->signature != MS_CBW_SIGNATURE) return 3;
	if(w->lun > MS_MAX_LUN) return 4;
	if(w->cb_length > 15) return 5;

	return 0;
}

/** 
 * Send a Mass Storage Control Status Wrapper and check it.
 *
 * @param w Where to read the CSW from.
 * @return Zero value on success, non-zero value on bus error.
 */
static int8_t ms_send_csw(const struct ms_csw *w)
{
	const uint8_t *p = (const uint8_t*)w;
	uint8_t i = sizeof(*w);

	usb_ep_select(USB_EPN_MS_IN);
	while(!usb_epfifo_ready()) {
		if(!usb_ep_enabled()) return -1;
	}
	while(i) {
		--i;
		usb_epfifo_push(*p++);
	}
	usb_in_clrbank();
	return 0;
}

/** 
 * Poll for Mass Storage Data requests.
 */
void ms_poll(void)
{
	/* Put the request and answer wrappers in same memory to save space */
	union {
		struct ms_cbw cbw;
		struct ms_csw csw;
	} u;
	enum ms_csw_status cmd_status;

	/* check whether we have received IN request for csw */
	usb_ep_select(USB_EPN_MS_IN);
	if(usb_ep_enabled() && usb_in_naksent()) {
		ms_stall_data_eps();
		return;
	}

	/* select the out endpoint and check whether we can receive a cbw */
	usb_ep_select(USB_EPN_MS_OUT);
	if(!usb_ep_enabled() || usb_ep_stall_enabled() || !usb_out_ready()) {
		return;
	}

	if(ms_recv_cbw(&u.cbw)) {
		ms_stall_data_eps();
		return;
	}

	/* Select endpoint for the data stage. */
	if(u.cbw.dir_d2h) {
		usb_ep_select(USB_EPN_MS_IN);
	} else {
		usb_ep_select(USB_EPN_MS_OUT);
	}
	/* Decode SCSI Block Command and execute the data stage */
	cmd_status = sbc_decode(u.cbw.cb, &u.cbw.data_transfer_length, u.cbw.dir_d2h);

	/* Execute the status stage. Tag has the same offset in CBW and CSW so
	 * don't touch it. The number of non-handled data bytes will be left
	 * in cbw.data_transfer_length. The latter has the same offset as
	 * csw.data_transfer_length so don't touch that one, either. 
	 *
	 * @note We may optimize for space even more by changing only the
	 * last signature byte.
	 */
	u.csw.signature = MS_CSW_SIGNATURE;
	/* csw.tag = cbw.tag; */
	/* csw.data_residue = cbw.data_transfer_length; */
	u.csw.status = cmd_status;
	if(ms_send_csw(&u.csw)) {
		ms_stall_data_eps();
	}
	usb_in_nakclr();
}

