/****************************************************************************
 *  
 * Project: Olimex AVR ISP with MKII protocol support.
 *
 * Copyright by Olimex Ltd. All rights reserved.
 * 
 * File: usb_core.c
 * Description: Core USB functions.
 * Developer: Dimitar Dimitrov ( dinuxbg  __on__  gmail.com )
 *
 * Last change: $Date: 2007-12-20 17:24:34 +0200 (четвъртък, 20 Декември 2007) $
 * Revision: $Revision: 97 $
 * Id: $Id: usb_core.c 97 2007-12-20 15:24:34Z dimitar $
 *
 ****************************************************************************/

#include <stdint.h>
#include <avr/pgmspace.h>

#include "usb_core.h"
#include "usb_conf.h"
#include "usb_hw.h"

uint8_t usb_current_configuration;

static void usb_get_status(const struct usb_setup_packet *stppkt);
static void usb_clear_feature(const struct usb_setup_packet *stppkt);
static void usb_set_feature(const struct usb_setup_packet *stppkt);
static void usb_set_address(const struct usb_setup_packet *stppkt);
static void usb_get_descriptor(const struct usb_setup_packet *stppkt);
static void usb_get_configuration(const struct usb_setup_packet *stppkt);
static void usb_set_configuration(const struct usb_setup_packet *stppkt);
static void usb_get_interface(const struct usb_setup_packet *stppkt);
static void usb_set_interface(const struct usb_setup_packet *stppkt);

static void usb_misc_request(const struct usb_setup_packet *stppkt)
{
	if(usbcb_setupreq(stppkt))
	{
		/* standard request or user-defined vendor or class request 
		 * not supported */
		usb_ep_stall_enable();
		usb_setupreq_sendack();
	}

}

/**
 * Handle incoming SETUP request. The whole setup packet is fetched an then
 * it is decoded. The function corresponding to each request type is called
 * to parse the setup packet and act on that. 
 *
 * @warning This function DOES NOT acknowledge the SETUP packet. Instead the
 * functions handling the request types do that because they might decide 
 * that the SETUP request is invalid and thus might stall endpoint zero!
 *
 * @warning This module assumes that there are no IN/OUT endpoint pairs 
 * sharing the same endpoint address! 
 */
void usb_core_setupreq(void)
{
	static struct usb_setup_packet stppkt;

	uint8_t i;
	uint8_t *p;

	/* fetch setup packet */
	p = (uint8_t*)&stppkt;
	for(i=0; i<sizeof(stppkt); ++i) {
		*p++ = usb_epfifo_pop();
	}

	if(stppkt.request_type.type != USB_REQUEST_TYPE_STANDARD) {
		usb_misc_request(&stppkt);
		return;
	}

	switch(stppkt.request) {
		case USB_STDREQ_GET_STATUS:
			if(!USB_STRICT_STANDARD || (stppkt.length==2
					&& stppkt.request_type.dir_d2h))
			{
				usb_get_status(&stppkt);
			} else {
				usb_misc_request(&stppkt);
			}
			break;
		case USB_STDREQ_CLEAR_FEATURE:
			if(!USB_STRICT_STANDARD || (!stppkt.length
					&& !stppkt.request_type.dir_d2h))
			{
				usb_clear_feature(&stppkt);
			} else {
				usb_misc_request(&stppkt);
			}
			break;
		case USB_STDREQ_SET_FEATURE:
			if(!USB_STRICT_STANDARD || (!stppkt.length
					&& !stppkt.request_type.dir_d2h))
			{
				usb_set_feature(&stppkt);
			} else {
				usb_misc_request(&stppkt);
			}
			break;
		case USB_STDREQ_SET_ADDRESS:
			if(!USB_STRICT_STANDARD || 
					(!stppkt.request_type.dir_d2h
					&& stppkt.request_type.recipient==USB_RECIPIENT_DEVICE
					&& !stppkt.index && !stppkt.length))
			{
				usb_set_address(&stppkt);
			} else {
				usb_misc_request(&stppkt);
			}
			break;
		case USB_STDREQ_GET_DESCRIPTOR:
			if(!USB_STRICT_STANDARD || (stppkt.request_type.dir_d2h
					&& (stppkt.request_type.recipient == USB_RECIPIENT_DEVICE)))
			{
				usb_get_descriptor(&stppkt);
			} else {
				usb_misc_request(&stppkt);
			}
			break;
		case USB_STDREQ_SET_DESCRIPTOR:
			usb_misc_request(&stppkt);
			break;
		case USB_STDREQ_GET_CONFIGURATION:
			if(!USB_STRICT_STANDARD || ((stppkt.length==1)
					&& (stppkt.request_type.recipient == USB_RECIPIENT_DEVICE)
					&& stppkt.request_type.dir_d2h))
			{
				usb_get_configuration(&stppkt);
			} else {
				usb_misc_request(&stppkt);
			}
			break;
		case USB_STDREQ_SET_CONFIGURATION:
			if(!USB_STRICT_STANDARD || ((stppkt.length==0)
					&& (stppkt.request_type.recipient == USB_RECIPIENT_DEVICE)
					&& !stppkt.request_type.dir_d2h))
			{
				usb_set_configuration(&stppkt);
			} else {
				usb_misc_request(&stppkt);
			}
			break;
		case USB_STDREQ_GET_INTERFACE:
			if(!USB_STRICT_STANDARD || ((stppkt.length==1)
					&& (stppkt.request_type.recipient == USB_RECIPIENT_INTERFACE)
					&& stppkt.request_type.dir_d2h))
			{
				usb_get_interface(&stppkt);
			} else {
				usb_misc_request(&stppkt);
			}
			break;
		case USB_STDREQ_SET_INTERFACE:
			if(!USB_STRICT_STANDARD || ((stppkt.length==0)
					&& (stppkt.request_type.recipient == USB_RECIPIENT_INTERFACE)
					&& !stppkt.request_type.dir_d2h))
			{
				usb_set_interface(&stppkt);
			} else {
				usb_misc_request(&stppkt);
			}
			break;
		case USB_STDREQ_SYNCH_FRAME:
			usb_misc_request(&stppkt);
			break;
		default:
			usb_misc_request(&stppkt);
	}
}


static void usb_get_status(const struct usb_setup_packet *stppkt)
{
	uint8_t ep_halted;

	switch((uint8_t)(stppkt->request_type.recipient)) {
		case USB_RECIPIENT_DEVICE:
			usb_setupreq_sendack();
			usb_epfifo_push(usb_is_self_powered()<<0);
			usb_epfifo_push(0x00);
			break;
		case USB_RECIPIENT_INTERFACE:
			/* FIXME - check stppkt.index */
			if(USB_STRICT_STANDARD) {
				if(!usb_dev_addressed()) {
					usb_ep_stall_enable();
					usb_setupreq_sendack();
					return;
				}
			}
			usb_setupreq_sendack();
			usb_epfifo_push(0x00);
			usb_epfifo_push(0x00);
			break;
		case USB_RECIPIENT_ENDPOINT:
			/* FIXME - check stppkt.index */
			if(USB_STRICT_STANDARD) {
				if(!usb_dev_addressed() && (stppkt->index&0xf)) 
				{
					usb_ep_stall_enable();
					usb_setupreq_sendack();
					return;
				}
			}
			usb_setupreq_sendack();
			usb_ep_select(stppkt->index & 0xf);
			ep_halted = usb_ep_stall_enabled() ? 1 : 0;
			usb_ep_select(USB_EPN_CONTROL);
			usb_epfifo_push(ep_halted);
			usb_epfifo_push(0x00);
			break;
		default:
			usb_ep_stall_enable();
			usb_setupreq_sendack();
			return;
	}

	usb_ctrlin_send();

	while(!usb_ctrlout_ready());
	usb_ctrlout_sendack();
}

static void usb_clear_feature(const struct usb_setup_packet *stppkt)
{
	if((stppkt->request_type.recipient==USB_RECIPIENT_ENDPOINT)
			&& (!USB_STRICT_STANDARD 
				|| stppkt->value==USB_STDFEAT_ENDPOINT_HALT))
	{
		if(USB_STRICT_STANDARD) {
			if(!usb_dev_addressed() && stppkt->index&0xf) {
				usb_ep_stall_enable();
				usb_setupreq_sendack();
				return;
			}
		}
		usb_ep_select(stppkt->index&0xf);
		if(usb_ep_enabled()) {
			if((uint8_t)stppkt->index != USB_EPN_CONTROL) {
				usb_ep_stall_disable();
				usb_ep_reset((uint8_t)(stppkt->index&0xf));
				usb_ep_reset_datatoggle();
			}
			usb_ep_select(stppkt->index & 0xf);
			usb_ep_stall_disable();
			usb_ep_select(USB_EPN_CONTROL);
			usb_setupreq_sendack();
			usb_ctrlin_send();
		} else {
			usb_ep_select(USB_EPN_CONTROL);
			usb_ep_stall_enable();
			usb_setupreq_sendack();
		}
	} else {
		usb_ep_stall_enable();
		usb_setupreq_sendack();
	}
}

static void usb_set_feature(const struct usb_setup_packet *stppkt)
{
	if((stppkt->request_type.recipient==USB_RECIPIENT_ENDPOINT)
			&& (!USB_STRICT_STANDARD
				|| stppkt->value==USB_STDFEAT_ENDPOINT_HALT))
	{
		if(USB_STRICT_STANDARD) {
			if(!usb_dev_addressed() && (stppkt->index&0xf)) {
				usb_ep_stall_enable();
				usb_setupreq_sendack();
				return;
			}
		}
		if((uint8_t)stppkt->index == USB_EPN_CONTROL) {
			usb_ep_stall_enable();
			usb_setupreq_sendack();
		} else {
			usb_ep_select((uint8_t)stppkt->index&0xf);
			if(usb_ep_enabled()) {
				usb_ep_stall_enable();
				usb_ep_select(USB_EPN_CONTROL);
				usb_setupreq_sendack();
				usb_ctrlin_send();
			} else {
				usb_ep_select(USB_EPN_CONTROL);
				usb_ep_stall_enable();
				usb_setupreq_sendack();
				usb_ctrlin_send();
			}
		}
	} else {
		usb_ep_stall_enable();
		usb_setupreq_sendack();
	}
}

static void usb_set_address(const struct usb_setup_packet *stppkt)
{
	if(!USB_STRICT_STANDARD || (~stppkt->value & 0x80)) {
		/* address good */
		usb_devaddr_load((uint8_t)stppkt->value);
		usb_setupreq_sendack();
		usb_ctrlin_send();
		while(!usb_ctrlin_ready());

		usb_devaddr_latch();
		usb_current_configuration = 0;
	} else {
		/* address beyond allowed range */
		usb_ep_stall_enable();
		usb_setupreq_sendack();
	}
}


/**
 * Handle request for fetching a descriptor.
 *
 * @warning The language ID for string descriptors is ignored!
 * @warning The descriptor index for configuration descriptors is ignored!
 */
static void usb_get_descriptor(const struct usb_setup_packet *stppkt)
{
	uint16_t descr_size;
	prog_uint8_t *descr_addr;
	uint8_t descr_index;

	descr_index = stppkt->value & 0xff;
	switch((uint8_t)(stppkt->value >> 8)) {
		case USB_DESCRTYPE_DEVICE:
			descr_size = sizeof(usb_device_descriptor);
			descr_addr = (prog_uint8_t*)&usb_device_descriptor;
			break;
		case USB_DESCRTYPE_CONFIGURATION:
			descr_size = sizeof(usb_configuration_block);
			descr_addr = (prog_uint8_t*)&usb_configuration_block;
			break;
		case USB_DESCRTYPE_STRING:
			if(descr_index < usb_get_strtbl_size()) {
				/* Warning: We believe that a pointer is 
				 * a 16bit variable! */
				descr_addr = (prog_uint8_t*)pgm_read_word(&usb_strings_table[descr_index]);
				descr_size = pgm_read_byte(&((struct usb_string_descriptor*)descr_addr)->length);
			} else {
				usb_ep_stall_enable();
				usb_setupreq_sendack();
				return;
			}
			break;
		case USB_DESCRTYPE_INTERFACE:
		case USB_DESCRTYPE_ENDPOINT:
			usb_ep_stall_enable();
			usb_setupreq_sendack();
			return;
			break;
		default:
			usb_ep_stall_enable();
			usb_setupreq_sendack();
			return;
	}
	
	usb_setupreq_sendack();
	usb_ctrlin_send_data(stppkt, descr_addr, descr_size, pgm_read_byte);
}

static void usb_get_configuration(const struct usb_setup_packet *stppkt)
{
	(void)stppkt;
	usb_setupreq_sendack();
	usb_epfifo_push(usb_current_configuration);
	usb_ctrlin_send();

	while(!usb_ctrlout_naksent());
	usb_ctrlout_nakclr();
	usb_ctrlout_sendack();
}

static void usb_set_configuration(const struct usb_setup_packet *stppkt)
{
	uint8_t cfg;

	cfg = stppkt->value;
	usb_current_configuration = cfg;
	
	if(!cfg) {
		/* reset all endpoints except the control one */
		for(uint8_t i=1; i<USB_EPN_TOTAL; ++i) {
			usb_ep_reset(i);
			usb_ep_select(i);
			usb_ep_disable();
			usb_ep_stall_disable();
			usb_epmem_deallocate();
		}

		usb_ep_select(USB_EPN_CONTROL);
		usb_setupreq_sendack();
	} else if(USB_STRICT_STANDARD && (cfg>1)) {
		/* we don't yet support more than one configuration */
		usb_ep_stall_enable();
		usb_setupreq_sendack();
		return;
	} else {
		usbcb_set_configuration();
		usb_ep_select(USB_EPN_CONTROL);
		usb_setupreq_sendack();
	}

	usb_ctrlin_send();
}

static void usb_get_interface(const struct usb_setup_packet *stppkt)
{
	(void)stppkt;
	if(USB_STRICT_STANDARD) {
		if(!usb_dev_addressed() || !usb_current_configuration) {
			usb_ep_stall_enable();
			usb_setupreq_sendack();
			return;
		}
	}
	usb_setupreq_sendack();
	usb_epfifo_push(0x00);	/* we support only one interface */
	usb_ctrlin_send();

	while(!usb_ctrlout_ready());
	usb_ctrlout_sendack();
}

static void usb_set_interface(const struct usb_setup_packet *stppkt)
{
	(void)stppkt;
	if(USB_STRICT_STANDARD) {
		if(!usb_dev_addressed() || !usb_current_configuration) {
			usb_ep_stall_enable();
			usb_setupreq_sendack();
			return;
		}
	}
	usb_setupreq_sendack();
	usb_ctrlin_send();
	while(!usb_ctrlin_ready());
}

